/* m-sun3.h -- 68020 version of the assembler for the Sun-3
   Copyright (C) 1987 Free Software Foundation, Inc.

This file is part of GAS, the GNU Assembler.

GAS is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GAS; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */



/* This header file contains the #defines specific
   to SUN computer SUN 3 series computers.  (The only kind
   we have around here, unfortunatly.)
   
   Rumor has it that this file will work on the Sun-2 if the assembler
   is called with -m68010  This is not tested.  */


#define M_SUN	1
#define SUN_ASM_SYNTAX
